#include "make_cert.h"

//    This file is part of MRCI_Client.

//    MRCI_Client is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.

//    MRCI_Client is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.

//    You should have received a copy of the GNU General Public License
//    along with MRCI_Client under the LICENSE.md file. If not, see
//    <http://www.gnu.org/licenses/>.

bool genDefaultSSLFiles(const QString &outsideAddr, QTextStream &msg)
{
    QProcess    proc;
    QStringList args;

    args << "req";
    args << "-newkey" << "rsa:4096";
    args << "-x509";
    args << "-sha256";
    args << "-days" << "365";
    args << "-nodes";
    args << "-out" << getLocalFilePath(DEFAULT_CERT_FILENAME);
    args << "-keyout" << getLocalFilePath(DEFAULT_PRIV_FILENAME);
    args << "-subj" << "/CN=" + outsideAddr;

    proc.setProgram("openssl");
    proc.setArguments(args);
    proc.start();

    msg << "openssl args: " << args.join(' ') << Qt::endl;

    auto ret = false;

    if (!proc.waitForStarted())
    {
        msg << "err: openssl failed to start, reason: " << proc.errorString() << Qt::endl;
    }
    else
    {
        proc.waitForFinished();

        msg << "--openssl-output--" << Qt::endl;

        auto outText = proc.readAllStandardOutput();
        auto errText = proc.readAllStandardError();

        msg << "std-out:" << Qt::endl;

        for (auto line : outText.split('\n'))
        {
            msg << line << Qt::endl;
        }

        msg << "std-err:" << Qt::endl;

        for (auto line: errText.split('\n'))
        {
            msg << line << Qt::endl;
        }

        if (proc.exitCode() != 0)
        {
            msg << "err: openssl failed with return code: " << proc.exitCode() << Qt::endl;
        }
        else
        {
            ret = true;
        }
    }

    return ret;
}
